#ifndef RV_CAMERA_H
#define RV_CAMERA_H

#include "RvDefs.h"
#include "RvGeom.h"
#include "RvTrackBall.h"
#include "RvSpace.h"

#include <assert.h>

class RvModel;
struct RvSpace;

#define NBVIEWS 5 // number of views stored in viewer

//////////////////////////////////////////////////////////
// frame rate control 
//////////////////////////////////////////////////////////

struct RvFrameRateCtlStruct
{
 	RvFrameRateCtlStruct()
	{
 	 RV_NB_CUR=0;
 	 RV_FRAME_RATE=0.0;
 	 RV_DT=0;
 	 RV_ODT=0;
 	 RV_OLDA=0;
 	 RV_FREQ=1.0;
	}

	void init(float FREQ=1.0);
	float getFrameRate() { return RV_FRAME_RATE;}
	void  enforce(int target=0);

protected:

 int RV_NB_CUR;
 double RV_FRAME_RATE;
 double RV_OLDT;
 double RV_OLDT2;
 double RV_DT;
 double RV_ODT;
 double RV_OLDA;
 double RV_FREQ;
};


/////////////////////////////////////////////////////////////
////// Camera Frustum

struct RvFrustum
{
	RvFrustum()
	{
		nearDist = 0.1;
		farDist = 100.0;
		fovAngle = RV_DEG2RAD(40.0);
		width =2.0;
		height =2.0;
		aspectRatio = 1.0;
	}

	void setSize(float w,float h) {
		aspectRatio = 1.0*w/h;
		width = w; height = h;
	}
	void setAspect(float aRatio)
	{ 
		aspectRatio = aRatio; 
		updateVPort();
	}
	void setAspect(float fovangle,float aRatio) 
	{
		fovAngle = fovangle;
		aspectRatio = aRatio;
		updateVPort();
	}



	float nearDist;
	float farDist;
protected:

	friend struct RvCamera;
	friend struct RvViewer;
	friend struct RvExaminerViewer;

	void updateFov()
	{ fovAngle = 2.0*atan2(height/2.0f,nearDist); }


	void updateVPort()
	{
		height = 2*nearDist*tan(fovAngle/2.0);
		width = aspectRatio*height;
		//cerr << "@@@@@@@@@@ "  << aspectRatio << "  " << fovAngle << " " << nearDist << "  " << width << "  " 
			//<< height << endl;
	}	

	float fovAngle;
	float aspectRatio;
	float width;
	float height;
};

//////////////////////////////////////////////////////////////////
/////// Camera 

#define RV_ORTHOGRAPHIC 0
#define RV_PERSPECTIVE 1

struct RvCamera
{
	RvCamera(int m = RV_PERSPECTIVE)
	{
		mode = m;
		position = RvVec(0,0,0);
		orientation = RvQuat::identity();
	}
	~RvCamera() {}

	void loadGL();

	void load(char *fname);
	void save(char *fname);

	void draw();

	void pointAt(const RvVec &target);
	void lookAt(const RvBox &target);

	void setPosition(const RvVec &pos) {position =pos;}
	const RvVec &getPosition() const {return position;}

	// Set orientation quaternion to orient
	void setOrientation(const RvQuat &orient) 
	{
		//...
	}

	void getGLMatrix(float *matrix);

	void rotate(const RvQuat &q)
	{
		orientation.applyQuat(q);
		RvVec p = position;
		q.applyTo(p);
		setPosition(p);
	}

	// return orientation -- uncomment
	//.....
//	const RvQuat &getOrientation() const { return orientation;}

	// convert orientation quaternion to up and towards vectors
	// dont forget to normalize
	void getOrientation(RvVec &twrds,RvVec &up) const 
	{
		// .....
		//cerr << twrds << " / " << up << endl;
	}

	RvFrustum & getFrustum() { return frust;}
	void setFrustum(const RvFrustum &f) {frust = f;}

	void toggleCameraType()
	{
		if (mode == RV_PERSPECTIVE)
			setCameraType(RV_ORTHOGRAPHIC);
		else
			setCameraType(RV_PERSPECTIVE);
	}

	void setCameraType(int m) { 
		if ((mode == RV_PERSPECTIVE) && (m==RV_ORTHOGRAPHIC))
		{
			// make sure width and height exist
			frust.updateVPort();
			mode = m;
		}
		else if ((m == RV_PERSPECTIVE) && (mode ==RV_ORTHOGRAPHIC))
		{
			// make sure fov exists
			frust.updateFov();
			mode = m;
		}
	}

protected:

	friend struct RvViewer;

	RvVec position;
	RvQuat orientation;

	RvFrustum frust;
	int mode;
};

//////////////////////////////////////////////////////////////////
/////// Viewers 

struct RvViewer
{
	RvViewer() { w = h= 0;
		zm = rot = trans = 0; full=0;
		fpsctl.init(1.0); 
		sel = 0; selection = NULL;
		dV = 0.0;
		curview = curviewrec = 0;
	}	
	
	~RvViewer() {}

	virtual void setScene(RvSpace *s) {space = s;}

	virtual void autoClipPlanes() {}

	virtual void drawAxes();

	virtual RvCamera &getCamera() { return cam;}
	virtual RvSpace *getScene() {return space;}

	virtual void setupRightView();
	virtual void setupLeftView();

	virtual void drawScene();

	virtual void readFrameBuffer(char *IMGBUF);

	// Need to fill this in
	virtual RvRay getRay(int mousex,int mousey)
	{
		RvRay r;
		//....
		return r;
	}

	virtual float getFrameRate() { return fpsctl.getFrameRate();}

	virtual void toggleRayPick() {
		if (!sel) sel = 1; else sel=0;
		selection = NULL; selpos = 0.0;
	}
	virtual int rayPickEnabled() { return sel;}

	virtual RvModel *getPrimitive(int mousex,int mousey,
		RvVec *pos = NULL, RvVec *norm = NULL, float *dist = NULL);
	virtual void rayPick(int mousex,int mousey)
	{ selection = getPrimitive(mousex,mousey, &selpos);}
	virtual RvModel *getSelection(){ return selection;}
	virtual RvVec &getSelectedLocation() { return selpos;}
		// ray pick functions

	virtual void resize(int width,int height)
	{
		w = width; h = height;
		cam.getFrustum().setAspect(1.0*w/h);
	}

	virtual void enableFullScreen() ;
	virtual void disableFullScreen();
	virtual void toggleFullScreen();
		// full screen functions

	virtual void saveCamera()
	{
		savepos[curview%NBVIEWS] = cam.getPosition();
		// Need to save orientation
		//....
		savefrust[curview%NBVIEWS] = cam.getFrustum();
		savemode[curview%NBVIEWS] = cam.mode;
		curview++;
	}
	virtual void recallCamera()
	{
		 if (!curview) return;
		 int NB = RV_MIN(curview,NBVIEWS);
		 cam.setPosition(savepos[curviewrec%NB]);
		 // Need to recover orientation
		//....
		 cam.setFrustum(savefrust[curviewrec%NB]);
		 cam.getFrustum().setAspect(1.0*w/h);
		 cam.mode = savemode[curviewrec%NB];
		 curviewrec++;
	}

	virtual void getWindowSize(int &x,int &y)
	{ x = w; y= h; }

	virtual void update(int mousex,int mousey);
		// to be called when mouse is moved

	virtual void refreshView(int targetFPS=0);
		// to be called in the display loop ONLY !!!

	virtual void processMouseEvents(int button,int button_state, 
				int modifiers, int x,int y);
		// to be called when a mouse button is pressed


protected:
	virtual void rotate(int mousex,int mousey) {}
	virtual void translate(int mousex,int mousey) {}
	virtual void zoom(int mousex,int mousey) {}

	RvCamera cam;
	RvSpace *space;

	RvVec savepos[NBVIEWS];
	RvQuat saveorient[NBVIEWS];
	RvFrustum savefrust[NBVIEWS];
	int savemode[NBVIEWS];
	int curview;
	int curviewrec;

	int w,h;
	int oldw, oldh;
	int full;
	// State variables for the operations
	// to be set in start/end
	int zm,rot,trans;
	int sel;

	RvFrameRateCtlStruct fpsctl;

	RvModel *selection;
	RvVec selpos;

	RvVec dV;

	float x,y;
	float startx,starty;
	RvVec T,U,W;
};


////////////////////////////////////////////////////////////

struct RvExaminerViewer : public RvViewer
{
	RvExaminerViewer() { 
		track.setRadius(0.7);
	}

	~RvExaminerViewer() {}

	void processMouseEvents(int button,int button_state, 
				int modifiers, int x,int y);

	virtual void recallCamera()
	{
		RvViewer::recallCamera();
		// Need to reset orientation
		//....
	}
	
protected:

	// this routine needs to initialize rotation parameters (quaternion)
	void beginRotate(int mousex,int mousey)
	{
		//.....
	}

	// this routine needs to finish up rotation parameters (quaternion)
	void endRotate() 
	{ 
		//....
	}
	// this routine needs to initialize translation parameters 
	void beginTranslate(int mousex,int mousey) 
	{ 
		//....
	}
	// this routine needs to finish up translation parameters 
	void endTranslate() 
	{ 
		//....
	}
	// same as above for zoom
	void beginZoom(int mousex,int mousey) 
	{ 	
		//....
	}
	void endZoom() 
	{
		//....
	}

	// perform the actual rotation (use quaternion)
	void rotate(int mousex,int mousey)
	{
		//....
		autoClipPlanes();
		cam.getFrustum().updateVPort();
	}

	// perform the actual translation 
	void translate(int mousex,int mousey)
	{

		//....
		autoClipPlanes();
		cam.getFrustum().updateVPort();
	}

	// perform zoom
	void zoom(int mousex,int mousey)
	{

		//....
		autoClipPlanes();
		cam.getFrustum().updateVPort();
	}

	virtual void autoClipPlanes() 
	{
		RvVec position = cam.getPosition();	
		RvBox target = space->bbox;
		float a = 
			(target.center() -position).length()
			- target.radius();

		if (a<=0.1) a=0.1;
		cam.getFrustum().nearDist = a;
		float b =
		cam.getFrustum().farDist  = a+2*target.radius()+1;

		//cerr << a <<  " " << b << endl;
	}


	// add fields for translation/rotate/soom
	RvVec startcampos;
	RvQuat Q;
	RvTrackBall track;
};


///////////////////////////////////////////////////////////

struct RvWalkViewer : public RvViewer
{
	RvWalkViewer() { speed = 0.1; rotspeed = 1.0; steer = 0.5; speedpersec= -1; rotspeedpersec = -1;}
	~RvWalkViewer() {}

	void processMouseEvents(int button,int button_state, 
				int modifiers, int x,int y);

	void refreshView(int fps = 0) 
	{
		RvViewer::refreshView(fps);
		float ff= getFrameRate();
		if (ff>1.0)
		{
			if (speedpersec>0.0) speed = ( (speed>=0) ? (speedpersec/ff) : -speedpersec/ff );	
			if (rotspeedpersec>0.0) rotspeed = rotspeedpersec/ff;	
		}
		if (trans)
		{
			RvVec V,U;
			// Need to set position and orientation
			// ....
		}
	}	

	void setSpeedMetersPerSec(float s) { speedpersec = s;}
	void setRotationSpeedDegPerSec(float s) { rotspeedpersec = s;}
	
	void setSpeed(float s) { speed = s;}
	void setRotationSpeed(float s) { rotspeed = s;}
	void setSteeringFactor(float f) { steer = f;}

protected:


	virtual void autoClipPlanes() 
	{
		RvVec position = cam.getPosition();	
		RvBox target = space->bbox;
		float a = 
			(target.center() -position).length()
			- target.radius();

		if (a<=0.1) a=0.1;
		cam.getFrustum().nearDist = a;
		float b =
		cam.getFrustum().farDist  = a+2*target.radius()+1;

		//cerr << a <<  " " << b << endl;
	}

	
	void beginTranslate(int mousex,int mousey)
	{ 
		//....
	}
	void endTranslate() 
	{ 
		//....
	}

	void beginRotate(int mousex,int mousey)
	{	
		//....
	}
	void endRotate() 
	{ 
		//....
	}
	void rotate(int mousex,int mousey)
	{
		//....
	}

	RvQuat QQ;
	float sfact;
	float speed;
	float rotspeed;
	float speedpersec;
	float rotspeedpersec;
	float steer;
};


//////////////////////////////////////////////////////////
/// Additional utility functions

void RvDrawText(int x,int y,char *txt,unsigned long rgba = RV_RGBA2INT(0,0,0,1));
void RvDraw3DArrow(const RvVec &origin,const RvVec &vec,
	unsigned long rgba, float scale =1.0,int n1=5,int n2=3);
void RvDrawDisc(const RvVec &c,const RvVec &n,float r,
unsigned long rgba,int n1=5,int n2=3);
void RvGLDisc(float r,int n1=5,int n2=3);
void RvGLCylinder(float r, float h,int n1=5,int n2=3);
void RvGLCone(float r, float h, int n1=5,int n2=3);
void RvGLCircle(float r,int n);
void	RvDrawCircle(const RvVec &c,const RvVec &n,float r,
		unsigned long rgba,int n1);

void RvDraw3DText(RvVec &pos,char *text,int color=RV_RGBA2INT(1,1,1,1), float scale=0.01);


#endif
