#ifndef RV_PRIMITIVE_H
#define RV_PRIMITIVE_H

#include "RvGeom.h"
#include "RvQuat.h"
// #include "RvProperties.h"

#define RV_POLYGON 1
#define RV_WEDGE 2
#define RV_PT 4
#define RV_ALL_PRIM_TYPE -1

struct RvPrimitive
{
	inline int type() {return _type;}

	virtual void draw(unsigned long rgba = RV_RGBA2INT(1,0,1,1),int drawaxes=1) =0;

	virtual int intersect(const RvRay &r,RvVec &pos,RvVec &norm,float &d,int backf_cull=1) = 0;
	virtual float distanceToPoint(const RvVec &v) {return 0.0;}

	virtual void scale(const RvVec &s) {}
	virtual void updateGeom(int glob) {}

	virtual const RvBox &getBbox() {return bbox;}

	/*
	virtual void setProperty(RvProperty *d) {prop = d;}
	virtual RvProperty *getProperty() { return prop;}
	*/

	virtual int isInside(const RvBox &box) {return 0;}

	virtual void show() { _show=1;}
	virtual void hide() { _show=0;}

	static RvPrimitive *getByName(char *name);
	virtual void setName(char *name);
	virtual const char *getName() { return name;}

	void *userData;

	int colorRGBA;

protected:
	RvPrimitive();

	int _type;

	char name [64];

	RvVec centroid;
	/*
	RvProperty *prop;
	*/

	RvBox bbox;
	int _show;

	RvPrimitive *next;
};

/////////////////////////////////////////////////

struct RvPt : public RvPrimitive
{
	RvPt(RvVec p =0.0);

    	void setPosition(float T,const RvVec &pos);
			// updates the state of the extrapolation mechanism
			// so it tends to reach the given position
	void  getPosition(float T,RvVec &pos);
			// returns the extrapolated position at time T
			// based on the current state of the extrapolation scheme
	void setOrientation(const RvQuat &q) { orient = q;}
	void getOrientation(RvQuat &q) { q= orient;}

	virtual void update(float time) { getPosition(time,pos); }
	virtual void getUpdatedPosition(RvVec &v) 
	//{ lock->lock(); v =pos; lock->unlock();}
	{  v =pos; }
	virtual void getRealPosition(RvVec &v) 
//	{ lock->lock(); 
	{	v = real;  }
// lock->unlock();}
	void enableExtrapolation() {extrap = 1;}
	void disableExtrapolation() { extrap = 0;}
	void tuneExtrapolation(float _minDt=1,float _dyna=4,float _dtpred=0.01) 
	{ minDt = _minDt; dyna=_dyna; dtpred = _dtpred;}

	float lastUpdateT() {return lupT;}	
		// last time the extrapolator state has been changed

	virtual void draw(unsigned long rgba = RV_RGBA2INT(1,0,1,1),int drawaxes=1);

	virtual int intersect(const RvRay &r,RvVec &POS,RvVec &norm,float&d,int backf_cull=1) { 
			RvVec v; getUpdatedPosition(v);
			if ((r.dir.dot(v-r.origin)>0) && (distanceToRay(r)<radius)){
				POS= v;
				d = (r.origin-v).length();
				return 1;
			}
			else return 0;
	}
	virtual float distanceToPoint(const RvVec &v) {return (pos-v).length();}

	virtual float distanceToRay(const RvRay &r)
	{
		RvVec v; getUpdatedPosition(v);
		RvVec v1 = v-r.origin;
		RvVec v2 = r.dir*(v1.dot(r.dir));
		//cerr << "toray " << (v1-v2).length() << endl;
		return (v1-v2).length();	
	}

	void setWidth(float w) {width = w;}
	float getWidth() { return width;}

	void getTowards(RvVec &vec);
	void getUp(RvVec &vec);
	void getRight(RvVec &vec);

	virtual int isInside(const RvBox &box) 
	{
		RvVec v; getUpdatedPosition(v);
		if ((v[0]>=box.min[0] && v[0]<=box.max[0])
		 && (v[1]>=box.min[1] && v[1]<=box.max[1])
		 && (v[2]>=box.min[2] && v[2]<=box.max[2]))
			return 1;
		else return 0;
	}


	float radius;

protected:
	int extrap;


        RvVec twrds;
        RvVec right;
        RvVec up;

	float minDt;
	float dyna;
	float dtpred;

	float width ;

	RvQuat orient;
	
	float lupT;

	/*
	RvLock *lock;
	*/

	RvVec oldReal;
	RvVec real;
	RvVec pos;

	RvVec A0,A,B,C;
	RvVec Q,QQ;
};

#define RV_MAX_SURF_VERTICES 8

/////////////////////////////////////////////////

struct RvSurface : public RvPrimitive
{
    RvSurface(int n, RvVec *vert,int lock=1);


	virtual void draw(unsigned long rgba = RV_RGBA2INT(1,0,1,1),int drawaxes=1);
	virtual int intersect(const RvRay &r,RvVec &pos,RvVec &norm,float &d,int backface_cull =1); 
	virtual float distanceToPoint(const RvVec &p);

	virtual void generateSamples(int nb,RvVec3* samples);

	virtual int contains(const RvVec &p);

	virtual void translate(RvVec &t) { _translation = t;}
	virtual void rotate(RvQuat &q) { _orient = q;}
	virtual void scale(const RvVec &s) { _scale = s;}

	void updateGeom(int global=0);	// use to update normal, edge dirs, area & plane params...
		// if global is true then the update replaces the current coordinates

	/*
	void lock() { if (_lock) _lock->lock();}
	void unlock() { if (_lock) _lock->unlock();}
	*/


	void setDisplayMaterial(RvDisplayMaterial *m) { material = m;}
	void setTextureCoords(float coords[][2])
	{
		for(int i=0;i<n;i++) 
		{ 
			texCoords[i][0] = coords[i][0];
			texCoords[i][1] = coords[i][1];
		}
	}

	float area;
    	RvVec norm;

    	int n;
    	RvVec vert[RV_MAX_SURF_VERTICES];
	RvVec VERT[RV_MAX_SURF_VERTICES];


protected:

	/*
	RvLock *_lock;
	*/

	RvVec _translation;
	RvVec _scale;
	RvQuat _orient;

	RvVec edges[RV_MAX_SURF_VERTICES];
	float A,B,C,D;

	float texCoords [RV_MAX_SURF_VERTICES][2];
	RvDisplayMaterial* material;
};

/////////////////////////////////////////////////

struct RvWedge : public RvPrimitive
{
	RvWedge(const RvVec &p1, const RvVec &p2,RvSurface *_s1,RvSurface *_s2)
	{
		edge[0]=p1; edge[1]=p2; s1=_s1; s2=_s2;
		width = 0.1;
		_type = RV_WEDGE;
	}

	virtual void draw(unsigned long rgba = RV_RGBA2INT(1,0,1,1),int drawaxes=1);
        virtual int intersect(const RvRay &r,RvVec &pos,RvVec &norm,float &d,int backface_cull =1) {return 0;}
		// intersect with a cylinder of radius width around the
		// edge of the wedge
        virtual float distanceToPoint(const RvVec &p) {return 0.0;}

	float width;

	RvVec edge[2];
	RvSurface *s1;
	RvSurface *s2;

};

////////////////////////////////////////////////////

#endif
