/*
******************************************************************************
**
** Functions definition
**
******************************************************************************
*/
#include <assert.h>
//#include <iostream>
#include "ff.h"
#include <string.h>
#include <stdio.h>
#include "objMtlParser.h"

int yy1parse();
int yymtlparse();

extern ffVertex gVertex;
extern ffNormal gNormal;
extern Vertices gVertices;
extern TextureCoord gTextureCoord;
extern Features gFeatures;
int gTriangleVertexIndex[3];

extern float gAmbientR;
extern float gAmbientG;
extern float gAmbientB;
extern bool  gHasAmbient;

extern float gDiffuseR;
extern float gDiffuseG;
extern float gDiffuseB;
extern bool  gHasDiffuse;

extern float gSpecularR;
extern float gSpecularG;
extern float gSpecularB;
extern bool  gHasSpecular;

extern float gSpecularN;
extern bool  gHasSpecularN;

static char textureFileName[1024];
static bool hasTexture=false;


/* basic structures */

/* store an re-allocatable array of vertices/normals/textcoords */

#define BLOCK_SIZE 25000

template  <class T> class mtlStorageClass  {
public:
	void Init();
	void Realloc();
	void Add(const T&);
	T& operator[](int i) { return stored_[i]; }
	int numStored() 	{	return numStored_; }
	T* stored() 	{	return stored_; }

	int init_;
	int numStored_, numReallocs_;
	T * stored_;
	mtlStorageClass() {
		stored_ = NULL;
		numStored_ = numReallocs_ = 0;
		init_ = 0;
	}
};

/* store an re-allocatable array of faces */
class ffFace {
public:
	int nverts;
	int indices[6]; // MAX OF 6 VERTS
	int materialID;
};

class ffMaterialFaceGroup {
public:
	int startFaceIndex, endFaceIndex;
};

static int currentFaceIndex=0;


static mtlStorageClass<ffVertex> storedVerts;
static mtlStorageClass<ffFace> storedFaces;
static mtlStorageClass<ffNormal> storedNormals;
static mtlStorageClass<TextureCoord> storedTextCoord;
static mtlStorageClass<objMtlSimpleMaterial> storedMaterials;
static mtlStorageClass<ffMaterialFaceGroup> storedMaterialFaceGroup;

static int 	objMtlCurrFace;

//--------------------------------------------------
// External interface


objMtlSimpleMaterial*
objMtlGetMaterialTable(int &n)
{
	n = storedMaterials.numStored();
	return storedMaterials.stored();
}

int objMtlReadModelFile( const char* szFileName, bool report )
{
	// reset for each file/object
	objMtlCurrFace = 0;
//	std::cerr << "Reading Model File " << szFileName << std::endl;
	printf("Reading Model File %s\n", szFileName);
	gVertices.maxIndex = 4;
	gVertices.xyzi = new int[gVertices.maxIndex];
	gVertices.uvi = new int[gVertices.maxIndex];
	gVertices.ni = new int[gVertices.maxIndex];

	if( !yy1Init( szFileName ) )
		return 0;
	yy1parse();
	yy1End();
	return 1;
}

int objMtlNumPolys()
{
	return storedFaces.numStored();
}

void objMtlGetVertices(unsigned int& numverts, objMtlSimpleVertex*& verts)
{
	numverts = storedVerts.numStored();
	verts = new objMtlSimpleVertex[numverts];
	for(int i=0; i<numverts; i++) {
		verts[i].xyz[0] = storedVerts[i].xyz[0];
		verts[i].xyz[1] = storedVerts[i].xyz[1];
		verts[i].xyz[2] = storedVerts[i].xyz[2];

		// check first
		if ( storedVerts[i].nindex != -1 ) {
			verts[i].n[0] = storedNormals[storedVerts[i].nindex].xyz[0];
			verts[i].n[1] = storedNormals[storedVerts[i].nindex].xyz[1];
			verts[i].n[2] = storedNormals[storedVerts[i].nindex].xyz[2];
		}

		if ( storedVerts[i].tindex != -1 ) {
			verts[i].uvw[0] = storedTextCoord[storedVerts[i].tindex].uvw[0];
			verts[i].uvw[1] = storedTextCoord[storedVerts[i].tindex].uvw[1];
//			verts[i].uvw[2] = storedTextCoord[storedVerts[i].tindex].uvw[2];
		}
	}
}

void objMtlGetIndices(int mesh, unsigned int& numindices, unsigned short int*& indices, int& materialID)
{
	int st = storedMaterialFaceGroup[mesh].startFaceIndex;
	int en = storedMaterialFaceGroup[mesh].endFaceIndex;
	numindices = (en-st)*3;
	materialID = storedFaces[st].materialID;

	indices = new unsigned short int[numindices];
	for(int i=st, j=0; i<en; i++) {
		if( storedFaces[i].nverts == 3 ) {
			indices[j++] = storedFaces[i].indices[0];
			indices[j++] = storedFaces[i].indices[1];
			indices[j++] = storedFaces[i].indices[2];
		}
//		else 
//			std::cerr << "Not a triangle, ignoring " << std::endl;
	}
}


//--------------------------------------------------

template <class T>
void
mtlStorageClass<T>::Init()
{
	init_ = true;
	numStored_ = numReallocs_ = 0;
	stored_ = new T [BLOCK_SIZE];
}

template <class T>
void
mtlStorageClass<T>::Realloc()
{
//	std::cerr << "Reallocing " << numStoredVerts << std::endl;
	T *tmp = stored_;
	numReallocs_++;

	stored_ = new T [(numReallocs_+1)*BLOCK_SIZE];
	for(int i=0; i<numStored_; i++) 
		stored_[i] = tmp[i];
	delete tmp;
}

template <class T>
void
mtlStorageClass<T>::Add(const T& elem)
{
	if( !init_ )
		Init();
	if( numStored_  >= (numReallocs_+1)*BLOCK_SIZE )
		Realloc();
	stored_[numStored_] = elem;
	numStored_++;

}

void AddVertex( ffVertex v )
{
	storedVerts.Add(v);
	v.index = 0;
}

void AddFace( int* vertices, int numVertices )
{
	ffFace faceToAdd;
	faceToAdd.nverts = numVertices;
	// Max vertices
	assert(numVertices<6);
	for(int i=0; i<numVertices; i++ ) 
		faceToAdd.indices[i] = vertices[i]-1;

	faceToAdd.materialID = gCurrentShader;
	storedFaces.Add(faceToAdd);

	/*
	std::cerr << "FACE " << std::endl;
	for(i=0; i<numVertices; i++) {
		std::cerr << "v " 
				<<  storedVerts[vertices[i]].xyz[0]  << " " 
				<<  storedVerts[vertices[i]].xyz[1]  << " " 
				<<  storedVerts[vertices[i]].xyz[2]  << std::endl;
	}
	*/
	currentFaceIndex ++;
}

void AddFace( int* vertices, int* textcoords, int* normals, int numVertices )
{
	ffFace faceToAdd;
	faceToAdd.nverts = numVertices;
	// Max vertices
	assert(numVertices<6);
	for(int i=0; i<numVertices; i++ ) {
		faceToAdd.indices[i] = vertices[i]-1;
		storedVerts[ vertices[i]-1 ].nindex = normals[i]-1;
		storedVerts[ vertices[i]-1 ].tindex = textcoords[i]-1;
	}
	faceToAdd.materialID = gCurrentShader;

	storedFaces.Add(faceToAdd);
	currentFaceIndex ++;
}

void AddFace( int* vertices, int* normals, int numVertices )
{
	ffFace faceToAdd;
	faceToAdd.nverts = numVertices;
	// Max vertices
	assert(numVertices<6);
	for(int i=0; i<numVertices; i++ ) {
		faceToAdd.indices[i] = vertices[i]-1;
		storedVerts[ vertices[i]-1 ].nindex = normals[i]-1;
	}

	faceToAdd.materialID = gCurrentShader;
	storedFaces.Add(faceToAdd);

	/*
	std::cerr << "FACE " << std::endl;
	for(i=0; i<numVertices; i++) {
		std::cerr << "v ( " << vertices[i]-1 << " ) "  
				<<  storedVerts[vertices[i]-1].xyz[0]  << " " 
				<<  storedVerts[vertices[i]-1].xyz[1]  << " " 
				<<  storedVerts[vertices[i]-1].xyz[2]  << std::endl;
	}
	*/
	currentFaceIndex ++;
}

/*! Reset the vertex information. */
void ResetVertexInfo()
{
  gVertex.index = 0;
}


/*! Reset the normal information. */
void ResetNormalInfo()
{
  gNormal.index = 0;
}


/*! Resets the vertices information. */
void ResetVerticesInfo()
{
  gVertices.index = 0;
  gVertices.usedUV = false;
  gVertices.usedN = false;
}


/*!
  Reset the texture coordinate information.
*/
void ResetTextureCoordinateInfo()
{
  gTextureCoord.index = 0;
}


/*!
  Resets the vertex information used when converting files (vertex
  compression).
*/
void ResetSavedInfo()
{
}


void
ResetMaterialInfo()
{
  gHasAmbient = false;
  gHasDiffuse = false;
  gHasSpecular = false;
  gHasSpecularN = false;
}

// called when you get a usemtl
int GetShader( const char* s)
{
	ffMaterialFaceGroup mfg;

	if(  storedMaterialFaceGroup.numStored() > 0 ) 
		storedMaterialFaceGroup[storedMaterialFaceGroup.numStored()-1].
			endFaceIndex = currentFaceIndex;


	mfg.startFaceIndex = currentFaceIndex;
	storedMaterialFaceGroup.Add(mfg);
	for(int i=0; i<storedMaterials.numStored(); i++) {
		if( !strcmp(storedMaterials[i].name, s))
			return i;
	}
	return -1;
}

void
CloseFaceGroup()
{
storedMaterialFaceGroup[storedMaterialFaceGroup.numStored()-1].endFaceIndex = currentFaceIndex;
}

int objMtlGetNumMeshes()
{
	return storedMaterialFaceGroup.numStored();
}

void
objMtlGetFaceGroupExtent(int i, int &st, int& en)
{
	st = storedMaterialFaceGroup[i].startFaceIndex;
	en = storedMaterialFaceGroup[i].endFaceIndex;
}



void AddNormal( ffNormal n )
{
	storedNormals.Add(n);
	n.index = 0;
}

void AddTextureCoordinate2d(  TextureCoord* t )
{
	TextureCoord toAdd = *t;
	/*
	std::cerr<< "Adding texture coordinate " 
			<< t->uvw[0] << " " 
			<< t->uvw[1] << " " 
			<< t->uvw[2] << std::endl;
	*/

	storedTextCoord.Add(toAdd);
	t->index = 0;
}


void AddColor( double r, double g, double b )
{
//	std::cerr << "Adding color "<< r << " " << g << " " << b << std::endl;
}

void AddDisplacementMap( /*enum TextureType*/ )
{
}


void AddShader( char* szMaterialName )
{
	objMtlSimpleMaterial mat;

	mat.name = szMaterialName;
	mat.Ka[0] = gAmbientR; mat.Ka[1] = gAmbientG; mat.Ka[2] = gAmbientB;
	mat.Kd[0] = gDiffuseR; mat.Kd[1] = gDiffuseG; mat.Kd[2] = gDiffuseB;
	mat.Ks[0] = gSpecularR; mat.Ks[1] = gSpecularG; mat.Ks[2] = gSpecularB;
	mat.Kt[0] = gTextureR; mat.Kt[1] = gTextureG; mat.Kt[2] = gTextureB;
	/*
	std::cerr << "ADDING SHADER " << szMaterialName << std::endl;

	std::cerr << "Ka " << mat.Ka[0] << " " << mat.Ka[1] << " " << mat.Ka[2]  << std::endl;
	std::cerr << "Kd " << mat.Kd[0] << " " << mat.Kd[1] << " " << mat.Kd[2]  << std::endl;
	*/
	if( hasTexture ) {
		// will be copied in Add
		mat.texture_name = textureFileName;
	}
	storedMaterials.Add(mat);
	hasTexture = false;
}

bool IsDefaultMaterialFile( const char* szName )
{
	return false;
}

void LoadTextureFile( const char* filename )
{
//	std::cerr << "Loading texture file " << filename << std::endl;
	printf("Loading texture file %s\n", filename);
	strcpy(textureFileName, filename);
	hasTexture = true;
}

void LoadMaterialFile( const char* filename )
{
//  std::cerr << "Loading Material File " << filename << std::endl;
    printf("Loading Material file %s\n", filename);
  if ( !yymtlInit( filename ) )
  {
    // try in data directory -- hack
    char path[1024];
    sprintf(path, "../../data/%s", filename);
     if( !yymtlInit(path)) {
	printf("[Error] ff.C::LoadMaterialFile : Could not initialize material parse with %s\n", filename);
//	  std::cerr << "[Error] ff.C::LoadMaterialFile : Could not initialize material "
 //        << "parser with " << filename << std::endl;
	    return;
    }
  }
  yymtlparse();
    yymtlEnd();

}

int convertModelFile( const char* szFileName, const char* szOutput )
{
	return true;
}

void ShutdownFileParser()
{
}

